(* ICFP Programming Contest 2001
 * Copyright (C) 2001 Jean-Christophe Filliatre *)

(*s The optimizer. *)

open Sml_ng
open Output
open Interp

(*s The optimized document is the list, in reverse order, of the
    following items. *)

type optimized_item = Char of char | Open of tag | Close of tag

type optimized_document = optimized_item list

(*s To output an optimized document, we reverse the list and then output the
    items; it could be done directly, but this way is tail recursive. *)

let output_optimized_document od =
  let output_item = function
    | Char c -> print_char c
    | Open t -> open_tag t
    | Close t -> close_tag t
  in
  List.iter output_item (List.rev od)

(*s The optimizer is based on the following function [switch], which
    finds out how to get the new context [ctx'] from the old one [ctx].
    [opens] is the stack of opened tags (in reverse order, the most recent
    coming first); they are paired with the context right before their opening.
    [acc] is the current optimized document (the list of 
    output items, in reverse order). 

    We proceed naively, dealing with the context features one by one
    (and thus [switch] is recursive).
    For properties to unset like \verb!B!, we close all the necessary tags,
    until the property disappear, using the function [close_until] (which
    takes a property to be satisfied as first argument).
    To set a property, we open the corresponding tag.
    For sizes and colors, we apply the following heuristic: if closing the
    last opened tag gives us the expected size (or color) we do that;
    otherwise, we open the corresponding tag. *)

let rec switch opens acc ctx ctx' =
  if ctx = ctx' then (* nothing to do *)
    opens, acc
  else if ctx.b && not ctx'.b then (* B to unset *)
    close_until (fun c -> not c.b) opens acc ctx'
  else if not ctx.b && ctx'.b then (* B to set *)
    switch ((B, ctx) :: opens) (Open B :: acc) { ctx with b = true } ctx'
  else if ctx.em && not ctx'.em then (* EM to unset *)
    close_until (fun c -> not c.em) opens acc ctx'
  else if not ctx.em && ctx'.em then (* EM to set *)
    switch ((EM, ctx) :: opens) (Open EM :: acc) { ctx with em = true } ctx'
  else if ctx.i && not ctx'.i then (* I to unset *)
    close_until (fun c -> not c.i) opens acc ctx'
  else if not ctx.i && ctx'.i then (* I to set *)
    switch ((I, ctx) :: opens) (Open I :: acc) { ctx with i = true } ctx'
  else if ctx.s && not ctx'.s then (* S to unset *)
    close_until (fun c -> not c.s) opens acc ctx'
  else if not ctx.s && ctx'.s then (* S to set *)
    switch ((S, ctx) :: opens) (Open S :: acc) { ctx with s = true } ctx'
  else if ctx.tt && not ctx'.tt then (* TT to unset *)
    close_until (fun c -> not c.tt) opens acc ctx'
  else if not ctx.tt && ctx'.tt then (* TT to set *)
    switch ((TT, ctx) :: opens) (Open TT :: acc) { ctx with tt = true } ctx'
  else if ctx.u > ctx'.u then (* U to decrease *)
    close_until (fun c -> c.u == ctx'.u) opens acc ctx'
  else if ctx.u < ctx'.u then (* U to increase *)
    switch ((U, ctx) :: opens) (Open U :: acc) { ctx with u = succ ctx.u } ctx'
  else if ctx.size != ctx'.size then (* Size to change *)
    let s = ctx'.size in (* desired size *)
    if s == -1 then
      close_until (fun c -> c.size == -1) opens acc ctx'
    else
      (match opens with
	 | (ti,ctxi) :: opens' when ctxi.size == s -> (* we're lucky *)
	     switch opens' (Close ti :: acc) ctxi ctx'
	 | _ ->
	     let t = Size s in
	     switch ((t, ctx) :: opens) (Open t :: acc) {ctx with size=s} ctx')
  else if ctx.color != ctx'.color then (* Color to change *)
    let c = ctx'.color in (* desired color *)
    if c == Any then
      close_until (fun c -> c.color == Any) opens acc ctx'
    else
      (match opens with
	 | (ti,ctxi) :: opens' when ctxi.color == c -> (* we're lucky *)
	     switch opens' (Close ti :: acc) ctxi ctx'
	 | _ ->
	     let t = Color c in
	     switch ((t, ctx)::opens) (Open t :: acc) {ctx with color=c} ctx')
  else (* necessarily [ctx = ctx'] *)
    assert false (* opens, acc *)

and close_until k opens acc ctx' = match opens with
  | [] -> 
      switch opens acc root_context ctx'
  | (ti,ctxi) :: opens' -> 
      if k ctxi then 
	switch opens' (Close ti :: acc) ctxi ctx'
      else
	close_until k opens' (Close ti :: acc) ctx'

(*s At the very end of the optimization, we close all the opened tags with
    the following function. *)

let rec close_all acc = function
  | [] -> acc
  | (t,_) :: opens -> close_all (Close t :: acc) opens

(*s Test for a space character. *)

let is_space = function
  | ' ' | '\t' | '\r' | '\n' -> true
  | _ -> false

(*s [update ctx ctx'] builds the context to reach from current context
    [ctx] when [ctx'] is the space context of the next character (thus a
    space). *)

let update ctx ctx' =
  { ctx with color = if ctx'.u > 0 then ctx'.color else ctx.color;
             u = ctx'.u; tt = ctx'.tt; size = ctx'.size }

(*s The optimizer itself. *)

let optim_document st =
  let rec loop opens acc ctx = parser
    | [< '(c,ctx'); st >] ->
	let ctx' = if is_space c then update ctx ctx' else ctx' in
	if ctx' = ctx then
	  loop opens (Char c :: acc) ctx st
	else
	  let opens', acc' = switch opens acc ctx ctx' in
	  loop opens' (Char c :: acc') ctx' st
    | [< >] -> 
	close_all acc opens
  in
  loop [] [] root_context st

(*s Size of the optimized document. *)

let size_item n = function
  | Char _ -> succ n
  | Open (EM | PL | TT) -> n + 4
  | Open _ -> n + 3
  | Close (EM | PL | TT) -> n + 5
  | Close _ -> n + 4

let size od = List.fold_left size_item 0 od

